#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <caml/alloc.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/custom.h>
#include <caml/fail.h>
#include <arpa/inet.h>

#include <time.h> // Only needed to avoid warning
#include "secp256k1.h"
#include "scalar_impl.h"
#include "testrand_impl.h"
#include "hash_impl.h"
#include "field_impl.h"
#include "group_impl.h"
#include "ecmult_const_impl.h"
#include "scratch_impl.h"

#include "ripemd160.h"

static struct custom_operations btc_ops = {
  (char*)"btc",
  custom_finalize_default,
  custom_compare_default,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default,
  custom_compare_ext_default
};

#define Btc_val(v) ((secp256k1_fe *)Data_custom_val(v))
static value alloc_btc() {
  return (alloc_custom(&btc_ops, sizeof(secp256k1_fe), 0, 1));
}

value c_sqrt_p(value x) {
  CAMLparam1 (x);
  CAMLlocal1(r);
  r = alloc_btc();
  secp256k1_fe_sqrt(Btc_val(r), Btc_val(x));
  secp256k1_fe_normalize(Btc_val(r));
  CAMLreturn(r);
}

value c_mul_p(value x, value y) {
  CAMLparam2 (x, y);
  CAMLlocal1(z);
  z = alloc_btc();
  if (Btc_val(x) == Btc_val(y)) {
    secp256k1_fe_sqr(Btc_val(z), Btc_val(x));
  } else {
    secp256k1_fe_mul(Btc_val(z), Btc_val(x), Btc_val(y));
  }
  secp256k1_fe_normalize(Btc_val(z));
  CAMLreturn (z);
}

value c_add_p(value x, value y) {
  CAMLparam2 (x, y);
  CAMLlocal1(z);
  z = alloc_btc();
  memcpy(Btc_val(z), Btc_val(x), sizeof(secp256k1_fe));
  secp256k1_fe_add(Btc_val(z), Btc_val(y));
  secp256k1_fe_normalize(Btc_val(z));
  CAMLreturn (z);
}

value c_sub_p(value x, value y) {
  CAMLparam2 (x, y);
  CAMLlocal1(z);
  z = alloc_btc();
  secp256k1_fe_negate(Btc_val(z), Btc_val(y), 1);
  secp256k1_fe_add(Btc_val(z), Btc_val(x));
  secp256k1_fe_normalize(Btc_val(z));
  CAMLreturn (z);
}

value c_isodd(value x) {
  CAMLparam1 (x);
  CAMLreturn(Val_bool(secp256k1_fe_is_odd(Btc_val(x))));
}

value c_iszero(value x) {
  CAMLparam1 (x);
  CAMLreturn(Val_bool(secp256k1_fe_normalizes_to_zero(Btc_val(x))));
}

value c_eq(value x,value y) {
  CAMLparam2 (x,y);
  CAMLreturn(Val_bool(secp256k1_fe_equal(Btc_val(x),Btc_val(y))));
}

value c_cmp(value x,value y) {
  CAMLparam2 (x,y);
  CAMLreturn(Val_int(secp256k1_fe_cmp_var(Btc_val(x),Btc_val(y))));
}

value c_neg_p(value x) {
  CAMLparam1 (x);
  CAMLlocal1(r);
  r = alloc_btc();
  secp256k1_fe_negate(Btc_val(r), Btc_val(x), 1);
  secp256k1_fe_normalize(Btc_val(r));
  CAMLreturn(r);
}

value c_inv_n(value x) {
  CAMLparam1 (x);
  unsigned char xxx[32];
  secp256k1_fe_get_b32(xxx, Btc_val(x));
  secp256k1_scalar xx1, xx2;
  secp256k1_scalar_set_b32(&xx1, xxx, NULL);
  secp256k1_scalar_inverse(&xx2, &xx1);
  secp256k1_scalar_get_b32(xxx, &xx2);
  CAMLlocal1(r);
  r = alloc_btc();
  secp256k1_fe_set_b32(Btc_val(r), xxx);
  CAMLreturn(r);
}

value c_inv_p(value x) {
  CAMLparam1 (x);
  CAMLlocal1(r);
  r = alloc_btc();
  secp256k1_fe_inv(Btc_val(r), Btc_val(x));
  secp256k1_fe_normalize(Btc_val(r));
  CAMLreturn(r);
}

value c_shr(value x, value n) {
  CAMLparam2(x,n);
  unsigned char xxx[32];
  secp256k1_fe_get_b32(xxx, Btc_val(x));
  secp256k1_scalar xx1;
  secp256k1_scalar_set_b32(&xx1, xxx, NULL);
  secp256k1_scalar_shr_int(&xx1, Long_val(n));
  secp256k1_scalar_get_b32(xxx, &xx1);
  CAMLlocal1(r);
  r = alloc_btc();
  secp256k1_fe_set_b32(Btc_val(r), xxx);
  CAMLreturn(r);
}

value c_of_small_int(value i){
  CAMLparam1(i);
  CAMLlocal1(o);
  o = alloc_btc();
  secp256k1_fe_set_int(Btc_val(o), Long_val(i));
  CAMLreturn(o);
}

value c_to_int32s(value s){
  CAMLparam1(s);
  unsigned char xxx[32];
  secp256k1_fe_get_b32(xxx, Btc_val(s));
  secp256k1_scalar xx;
  secp256k1_scalar_set_b32(&xx, xxx, NULL);
  CAMLlocal4(a,b,c,d);
  a = caml_copy_int32(((int32_t*)xx.d)[7]);
  b = caml_copy_int32(((int32_t*)xx.d)[6]);
  c = caml_copy_int32(((int32_t*)xx.d)[5]);
  d = caml_copy_int32(((int32_t*)xx.d)[4]);
  CAMLlocal4(e,f,g,h);
  e = caml_copy_int32(((int32_t*)xx.d)[3]);
  f = caml_copy_int32(((int32_t*)xx.d)[2]);
  g = caml_copy_int32(((int32_t*)xx.d)[1]);
  h = caml_copy_int32(((int32_t*)xx.d)[0]);
  CAMLlocal1 (r);
  r = caml_alloc_tuple(8);
  Field(r, 0) = a;
  Field(r, 1) = b;
  Field(r, 2) = c;
  Field(r, 3) = d;
  Field(r, 4) = e;
  Field(r, 5) = f;
  Field(r, 6) = g;
  Field(r, 7) = h;
  CAMLreturn (r);
}

value c_of_int32s(value ox){
  CAMLparam1(ox);
  secp256k1_scalar xx;
  ((uint32_t*)(xx.d))[0] = Int32_val(Field(ox, 7));
  ((uint32_t*)(xx.d))[1] = Int32_val(Field(ox, 6));
  ((uint32_t*)(xx.d))[2] = Int32_val(Field(ox, 5));
  ((uint32_t*)(xx.d))[3] = Int32_val(Field(ox, 4));
  ((uint32_t*)(xx.d))[4] = Int32_val(Field(ox, 3));
  ((uint32_t*)(xx.d))[5] = Int32_val(Field(ox, 2));
  ((uint32_t*)(xx.d))[6] = Int32_val(Field(ox, 1));
  ((uint32_t*)(xx.d))[7] = Int32_val(Field(ox, 0));
  unsigned char xxx[32];
  secp256k1_scalar_get_b32(xxx, &xx);
  CAMLlocal1(o);
  o = alloc_btc();
  secp256k1_fe_set_b32(Btc_val(o), xxx);
  CAMLreturn(o);
}

value c_to_lebits(value x){
  CAMLparam1(x);
  CAMLlocal1(r);
  r = caml_alloc_string(32);
  unsigned char tmp;
  secp256k1_fe_get_b32((unsigned char*)(String_val(r)), Btc_val(x));
  for (int i = 0; i < 16; ++i) {
    tmp=Byte_u(r, i); Byte_u(r, i) = Byte_u(r, 31 - i); Byte_u(r, 31 - i) = tmp;
  }
  CAMLreturn(r);
}

value c_of_lebits(value x){
  CAMLparam1(x);
  CAMLlocal1(r);
  r = alloc_btc();
  unsigned char xxx[32];
  memset(xxx, 0, 32);
  memcpy(xxx, String_val(x), caml_string_length(x));
  unsigned char tmp;
  for (int i = 0; i < 16; ++i) {
    tmp=xxx[i]; xxx[i]=xxx[31 - i]; xxx[31 - i]=tmp;
  }
  secp256k1_fe_set_b32(Btc_val(r), xxx);
  CAMLreturn(r);
}

value c_to_bebits(value x){
  CAMLparam1(x);
  CAMLlocal1(r);
  r = caml_alloc_string(32);
  secp256k1_fe_get_b32((unsigned char*)(String_val(r)), Btc_val(x));
  CAMLreturn(r);
}

value c_of_bebits(value x){
  CAMLparam1(x);
  CAMLlocal1(r);
  r = alloc_btc();
  unsigned char xxx[32];
  memset(xxx, 0, 32);
  const mlsize_t l = caml_string_length(x);
  memcpy(xxx + 32 - l, String_val(x), l);
  secp256k1_fe_set_b32(Btc_val(r), xxx);
  CAMLreturn(r);
}

static struct custom_operations ge_ops = {
  (char*)"ge",
  custom_finalize_default,
  custom_compare_default,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default,
  custom_compare_ext_default
};

#define Ge_val(v) ((secp256k1_ge *)Data_custom_val(v))
static value alloc_ge() {
  return (alloc_custom(&ge_ops, sizeof(secp256k1_ge), 0, 1));
}

value c_curve_y(value odd, value x){
  CAMLparam2 (odd,x);
  CAMLlocal1(r);
  r = alloc_ge();
  secp256k1_ge_set_xo_var(Ge_val(r), Btc_val(x), Bool_val(odd));
  CAMLreturn(r);
}


value c_ge_of_xy(value x, value y){
  CAMLparam2 (x,y);
  CAMLlocal1(r);
  r = alloc_ge();
  secp256k1_ge_set_xy(Ge_val(r), Btc_val(x), Btc_val(y));
  CAMLreturn(r);
}

value c_ge_to_xyo(value g){
  CAMLparam1 (g);
  if (Ge_val(g)->infinity) CAMLreturn(Val_unit);
  CAMLlocal4(r,t,x,y);
  x = alloc_btc();
  memcpy(Btc_val(x), &Ge_val(g)->x, sizeof(secp256k1_fe));
  secp256k1_fe_normalize(Btc_val(x));
  y = alloc_btc();
  memcpy(Btc_val(y), &Ge_val(g)->y, sizeof(secp256k1_fe));
  secp256k1_fe_normalize(Btc_val(y));
  t = caml_alloc_tuple(2);
  Field(t, 0) = x;
  Field(t, 1) = y;
  r = caml_alloc(1, 0); //Val_some(t);
  Field(r, 0) = t;
  CAMLreturn(r);
}

value c_smulp(value s, value p){
  CAMLparam2 (s,p);
  unsigned char xxx[32];
  //secp256k1_fe_normalize(Btc_val(s));
  secp256k1_fe_get_b32(xxx, Btc_val(s));
  secp256k1_scalar xx1;
  secp256k1_scalar_set_b32(&xx1, xxx, NULL);
  secp256k1_gej r;
  secp256k1_ecmult_const(&r, Ge_val(p), &xx1, 257);
  CAMLlocal1(o);
  o = alloc_ge();
  secp256k1_ge_set_gej(Ge_val(o), &r);
  CAMLreturn(o);
}

value c_add_pt(value x, value y){
  CAMLparam2(x,y);
  secp256k1_gej rj,xj,yj;
  secp256k1_gej_set_ge(&xj, Ge_val(x));
  secp256k1_gej_set_ge(&yj, Ge_val(y));
  secp256k1_gej_add_var(&rj, &xj, &yj, NULL);
  CAMLlocal1(r);
  r = alloc_ge();
  secp256k1_ge_set_gej(Ge_val(r), &rj);
  CAMLreturn(r);
}

value c_sha256_bebits(value s){
  CAMLparam1(s);
  secp256k1_sha256 hasher;
  secp256k1_sha256_initialize(&hasher);
  secp256k1_sha256_write(&hasher, (const unsigned char*)(String_val(s)), caml_string_length(s));
  CAMLlocal1(r);
  r = caml_alloc_string(32);
  secp256k1_sha256_finalize(&hasher, (unsigned char*)(String_val(r)));
  CAMLreturn(r);
}

value c_sha256_round(value chv, value cb){
  CAMLparam2(chv,cb);
  // Supposed to be big endian already
  int32_t s[8] = {Int32_val(Field(chv,0)),Int32_val(Field(chv,1)),Int32_val(Field(chv,2)),Int32_val(Field(chv,3)),
                   Int32_val(Field(chv,4)),Int32_val(Field(chv,5)),Int32_val(Field(chv,6)),Int32_val(Field(chv,7))};
  // In host order but library needs big
  int32_t cc[16] = {htonl(Int32_val(Field(cb,0))),htonl(Int32_val(Field(cb,1))),htonl(Int32_val(Field(cb,2))),htonl(Int32_val(Field(cb,3))),
                    htonl(Int32_val(Field(cb,4))),htonl(Int32_val(Field(cb,5))),htonl(Int32_val(Field(cb,6))),htonl(Int32_val(Field(cb,7))),
                    htonl(Int32_val(Field(cb,8))),htonl(Int32_val(Field(cb,9))),htonl(Int32_val(Field(cb,10))),htonl(Int32_val(Field(cb,11))),
                    htonl(Int32_val(Field(cb,12))),htonl(Int32_val(Field(cb,13))),htonl(Int32_val(Field(cb,14))),htonl(Int32_val(Field(cb,15)))};
  secp256k1_sha256_transform((uint32_t*)s, (uint32_t*)cc);
  CAMLlocal4(a,b,c,d);
  a = caml_copy_int32(s[0]);
  b = caml_copy_int32(s[1]);
  c = caml_copy_int32(s[2]);
  d = caml_copy_int32(s[3]);
  CAMLlocal4(e,f,g,h);
  e = caml_copy_int32(s[4]);
  f = caml_copy_int32(s[5]);
  g = caml_copy_int32(s[6]);
  h = caml_copy_int32(s[7]);
  CAMLlocal1 (r);
  r = caml_alloc_tuple(8);
  Field(r, 0) = a;
  Field(r, 1) = b;
  Field(r, 2) = c;
  Field(r, 3) = d;
  Field(r, 4) = e;
  Field(r, 5) = f;
  Field(r, 6) = g;
  Field(r, 7) = h;
  CAMLreturn (r);
}

value c_ripemd160_be_be(value s){
  CAMLparam1(s);
  if (caml_string_length(s) != 32)
    caml_invalid_argument("c_ripemd160_be_be : only accepts strings of length 32");
  CAMLlocal1 (r);
  r = caml_alloc_string(20);
  ripemd160_initialize((uint32_t*)((void*)(String_val(r))));
  int32_t cc[16];
  memcpy(cc, (String_val(s)), 8 * sizeof(uint32_t));
  memset(&cc[9], 0, 7 * sizeof(uint32_t)); cc[8] = 0x80; cc[14] = 256;
  ripemd160_transform((uint32_t*)((void*)(String_val(r))), (uint32_t*)cc);
  CAMLreturn (r);
}

/* /* Works, but we don't need it anymore.
value c_ripemd160_tuples(value cb){
  CAMLparam1(cb);
  int32_t s[5];
  ripemd160_initialize((uint32_t*)s);
  int32_t cc[16] = {htonl(Int32_val(Field(cb,0))),htonl(Int32_val(Field(cb,1))),htonl(Int32_val(Field(cb,2))),htonl(Int32_val(Field(cb,3))),
                    htonl(Int32_val(Field(cb,4))),htonl(Int32_val(Field(cb,5))),htonl(Int32_val(Field(cb,6))),htonl(Int32_val(Field(cb,7))),
                    0x80,0,0,0,0,0,256,0};
  ripemd160_transform((uint32_t*)s, (uint32_t*)cc);
  CAMLlocal5(a,b,c,d,e);
  a = caml_copy_int32(ntohl(s[0]));
  b = caml_copy_int32(ntohl(s[1]));
  c = caml_copy_int32(ntohl(s[2]));
  d = caml_copy_int32(ntohl(s[3]));
  e = caml_copy_int32(ntohl(s[4]));
  CAMLlocal1 (r);
  r = caml_alloc_tuple(5);
  Field(r, 0) = a;
  Field(r, 1) = b;
  Field(r, 2) = c;
  Field(r, 3) = d;
  Field(r, 4) = e;
  CAMLreturn (r);
}

value c_ripemd160_round(value chv, value cb){
  CAMLparam2(chv,cb);
  int32_t s[5] = {Int32_val(Field(chv,0)),Int32_val(Field(chv,1)),Int32_val(Field(chv,2)),Int32_val(Field(chv,3)), Int32_val(Field(chv,4))};
  int32_t cc[16] = {(Int32_val(Field(cb,0))),(Int32_val(Field(cb,1))),(Int32_val(Field(cb,2))),(Int32_val(Field(cb,3))),
                    (Int32_val(Field(cb,4))),(Int32_val(Field(cb,5))),(Int32_val(Field(cb,6))),(Int32_val(Field(cb,7))),
                    (Int32_val(Field(cb,8))),(Int32_val(Field(cb,9))),(Int32_val(Field(cb,10))),(Int32_val(Field(cb,11))),
                    (Int32_val(Field(cb,12))),(Int32_val(Field(cb,13))),(Int32_val(Field(cb,14))),(Int32_val(Field(cb,15)))};
  ripemd160_transform((uint32_t*)s, (uint32_t*)cc);
  CAMLlocal5(a,b,c,d,e);
  a = caml_copy_int32(s[0]);
  b = caml_copy_int32(s[1]);
  c = caml_copy_int32(s[2]);
  d = caml_copy_int32(s[3]);
  e = caml_copy_int32(s[4]);
  CAMLlocal1 (r);
  r = caml_alloc_tuple(5);
  Field(r, 0) = a;
  Field(r, 1) = b;
  Field(r, 2) = c;
  Field(r, 3) = d;
  Field(r, 4) = e;
  CAMLreturn (r);
}
*/
