(* Copyright (c) 2021 The Proofgold Lava developers *)
(* Copyright (c) 2020-2021 The Proofgold Core developers *)
(* Copyright (c) 2020 The Proofgold developers *)
(* Copyright (c) 2017-2018 The Dalilcoin developers *)
(* Distributed under the MIT software license, see the accompanying
   file COPYING or http://www.opensource.org/licenses/mit-license.php. *)

open Ser
open Hash

let dbdir = ref ""

exception NoBootstrapURL

let dbconfig dir =
  dbdir := dir;
  if Sys.file_exists dir then
    if Sys.is_directory dir then
      ()
    else
      raise (Failure (dir ^ " is a file not a directory"))
  else
    begin
      Unix.mkdir dir 0b111111000;
      Utils.forward_from_ltc_block := Some("26c874bca3122a06ec9b6582d4b62f38cfbf9a490da15698fe9a33c7d5a35cde")
    end

module type dbmtype = functor (M:sig type t val basedir : string val seicval : (seict -> t * seict) val seival : (seist -> t * seist) val seoval : (t -> seosbt -> seosbt) end) ->
  sig
    val dbinit : unit -> unit
    val dbget : hashval -> M.t
    val dbget_opt : hashval -> M.t option
    val dbexists : hashval -> bool
    val dbput : hashval -> M.t -> unit
    val dbdelete : hashval -> unit
    val dbkeyiter : (hashval -> unit) -> unit
    val dbbincp : hashval -> Buffer.t -> unit
  end
                            
module Dbmbasic : dbmtype = functor (M:sig type t val basedir : string val seicval : (seict -> t * seict) val seival : (seist -> t * seist) val seoval : (t -> seosbt -> seosbt) end) ->
  struct
    let mutexdb : Mutex.t = Mutex.create()

    let withlock f =
      try
	Mutex.lock mutexdb;
	let r = f() in
	Mutex.unlock mutexdb;
	r
      with e ->
	Mutex.unlock mutexdb;
	raise e

    let db1 = ref None
            
    let dbinit () =
      let d2 = Filename.concat !dbdir M.basedir in
      if Sys.file_exists d2 then
        (if not (Sys.is_directory d2) then raise (Failure (d2 ^ " is a file not a directory")))
      else
        Unix.mkdir d2 0b111111000;
      let d2b = Filename.concat d2 "db4" in
      let d4 = Mlgdbm.dbopen (d2b ^ ".pag") in
      db1 := Some(d4);
      let rec importdb d =
        if Sys.file_exists d && Sys.is_directory d then
          let di = Filename.concat d "index" in
          let dd = Filename.concat d "data" in
          let ci = open_in_bin di in
          let cd = open_in_bin dd in
          begin
            try
              while true do
                let (h,_) = sei_hashval seic (ci,None) in
	        let (p,_) = sei_int32 seic (ci,None) in
                seek_in cd (Int32.to_int p);
                let (v,_) = M.seicval (cd,None) in
                let k = Be256.to_string h in
                let s = Marshal.to_string v [Marshal.No_sharing;Marshal.Compat_32] in
                Mlgdbm.replace d4 k s
              done
            with End_of_file ->
              close_in ci;
              close_in cd
          end;
          List.iter
            (fun sd ->
              let dk = Filename.concat d sd in
              importdb dk)
            ["00";"01";"02";"03";"04";"05";"06";"07";"08";"09";"0a";"0b";"0c";"0d";"0e";"0f";
	     "10";"11";"12";"13";"14";"15";"16";"17";"18";"19";"1a";"1b";"1c";"1d";"1e";"1f";
	     "20";"21";"22";"23";"24";"25";"26";"27";"28";"29";"2a";"2b";"2c";"2d";"2e";"2f";
	     "30";"31";"32";"33";"34";"35";"36";"37";"38";"39";"3a";"3b";"3c";"3d";"3e";"3f";
	     "40";"41";"42";"43";"44";"45";"46";"47";"48";"49";"4a";"4b";"4c";"4d";"4e";"4f";
	     "50";"51";"52";"53";"54";"55";"56";"57";"58";"59";"5a";"5b";"5c";"5d";"5e";"5f";
	     "60";"61";"62";"63";"64";"65";"66";"67";"68";"69";"6a";"6b";"6c";"6d";"6e";"6f";
	     "70";"71";"72";"73";"74";"75";"76";"77";"78";"79";"7a";"7b";"7c";"7d";"7e";"7f";
	     "80";"81";"82";"83";"84";"85";"86";"87";"88";"89";"8a";"8b";"8c";"8d";"8e";"8f";
	     "90";"91";"92";"93";"94";"95";"96";"97";"98";"99";"9a";"9b";"9c";"9d";"9e";"9f";
	     "a0";"a1";"a2";"a3";"a4";"a5";"a6";"a7";"a8";"a9";"aa";"ab";"ac";"ad";"ae";"af";
	     "b0";"b1";"b2";"b3";"b4";"b5";"b6";"b7";"b8";"b9";"ba";"bb";"bc";"bd";"be";"bf";
	     "c0";"c1";"c2";"c3";"c4";"c5";"c6";"c7";"c8";"c9";"ca";"cb";"cc";"cd";"ce";"cf";
	     "d0";"d1";"d2";"d3";"d4";"d5";"d6";"d7";"d8";"d9";"da";"db";"dc";"dd";"de";"df";
	     "e0";"e1";"e2";"e3";"e4";"e5";"e6";"e7";"e8";"e9";"ea";"eb";"ec";"ed";"ee";"ef";
	     "f0";"f1";"f2";"f3";"f4";"f5";"f6";"f7";"f8";"f9";"fa";"fb";"fc";"fd";"fe";"ff"]
      in
      importdb (Filename.concat (Filename.concat !dbdir "../../.proofgoldcore/db") M.basedir)
      
    let dbf () =
      match !db1 with
      | Some(db) ->
         db
      | None ->
         let d2b = Filename.concat (Filename.concat !dbdir M.basedir) "db4" in
         let db = Mlgdbm.dbopen (d2b ^ ".pag") in
         db1 := Some(db);
         db

    let dbexists k =
      let k = Be256.to_string k in
      Mlgdbm.exists (dbf ()) k

    let dbget k =
      let k = Be256.to_string k in
      ((Mlgdbm.find_unmarshal (dbf ()) k) : M.t)

    let dbget_opt k =
      let k = Be256.to_string k in
      ((Mlgdbm.find_unmarshal_opt (dbf ()) k) : M.t option)

    let dbput k v =
      let k = Be256.to_string k in
      withlock
        (fun () ->
          let s = Marshal.to_string v [Marshal.No_sharing;Marshal.Compat_32] in
          Mlgdbm.replace (dbf()) k s)

    let dbdelete k =
      let k = Be256.to_string k in
      withlock
        (fun () -> Mlgdbm.delete (dbf ()) k)

    let dbkeyiter f =
      let db = dbf () in
      let rec walk = function
	| None -> ()
	| Some k ->
           f (Be256.of_substring k 0);
           walk (try Some (Mlgdbm.nextkey db k) with Not_found -> None)
      in
      walk (try Some(Mlgdbm.firstkey db) with Not_found -> None)

    let dbbincp k sb =
      let k = Be256.to_string k in
      let s = Mlgdbm.find (dbf ()) k in
      Buffer.add_string sb s

  end

module DbBlacklist = Dbmbasic (struct type t = bool let basedir = "blacklist" let seicval = sei_bool seic let seival = sei_bool seis let seoval = seo_bool seosb end)

module DbArchived = Dbmbasic (struct type t = bool let basedir = "archived" let seicval = sei_bool seic let seival = sei_bool seis let seoval = seo_bool seosb end)
